import json
from red_warden.config import RWConfig
from red_warden.datalayer import RWMysqlBackend
from red_warden.helpers import import_string
from red_warden.enums import RWBackendDestinationEnum


async def upgrade(db):
    async with RWMysqlBackend(
        RWConfig["RW_BACKEND_ENDPOINTS"],
        {
            "username": RWConfig["RW_BACKEND_USERNAME"],
            "password": RWConfig["RW_BACKEND_PASSWORD"],
        },
    ) as _db:
        # check if database exists, else creates it
        db_exists = await _db.fetch_val(
            "SHOW DATABASES LIKE :db_name", {"db_name": RWConfig["RW_BACKEND_DB_NAME"]}
        )
        if not db_exists:
            await _db.execute(
                "CREATE DATABASE %s DEFAULT CHARACTER SET utf8mb4 COLLATE utf8mb4_0900_ai_ci"
                % RWConfig["RW_BACKEND_DB_NAME"]
            )

            # ensure all grants are set for this user to the Manager database and tables
            await _db.execute(
                "GRANT ALL ON %s.* TO :username@'%%'" % RWConfig["RW_BACKEND_DB_NAME"],
                {"username": RWConfig["RW_BACKEND_USERNAME"]},
            )

    async with db:
        await db.execute(
            """        
            CREATE TABLE `backend` (
              `id` binary(16) NOT NULL,
              `name` varchar(250) NOT NULL,
              `type` enum('mysql','mongodb','redis') NOT NULL,
              `destination` enum('global', 'tenants_available', 'tenants_full') NOT NULL,
              `params` varchar(250) DEFAULT NULL,
              `endpoints` varchar(250) NOT NULL,
              `enabled` enum('Y','N') NOT NULL DEFAULT 'N',
              PRIMARY KEY (`id`),
              UNIQUE KEY `backend_name_uindex` (`name`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `tenant` (
              `id` binary(16) NOT NULL,
              `name` varchar(250) NOT NULL,
              `params` varchar(250) DEFAULT NULL,
              `enabled` enum('Y','N') NOT NULL DEFAULT 'N',
              PRIMARY KEY (`id`),
              UNIQUE KEY `tenant_name_uindex` (`name`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `datapath` (
              `id` binary(16) NOT NULL,
              `tenant_id` binary(16) DEFAULT NULL,
              `datazone` varchar(250) NOT NULL,
              `backend_id` binary(16) NOT NULL,
              `db_name` varchar(100) DEFAULT NULL,
              `version` bigint NOT NULL,
              PRIMARY KEY (`id`),
              KEY `datapath_backend_id_fk` (`backend_id`),
              KEY `datapath_tenant_id_fk` (`tenant_id`),
              CONSTRAINT `datapath_backend_id_fk` FOREIGN KEY (`backend_id`) REFERENCES `backend` (`id`) ON DELETE CASCADE,
              CONSTRAINT `datapath_tenant_id_fk` FOREIGN KEY (`tenant_id`) REFERENCES `tenant` (`id`) ON DELETE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `key` (
              `id` binary(16) NOT NULL,
              `login` varchar(250) NOT NULL,
              `password_hash` varchar(250) NOT NULL,
              `enabled` enum('Y','N') NOT NULL DEFAULT 'N',
              `mfa_enabled` enum('Y','N') NOT NULL DEFAULT 'N',
              `mfa_secret` varchar(250) DEFAULT NULL,
              PRIMARY KEY (`id`),
              UNIQUE KEY `key_login_uindex` (`login`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `session` (
              `id` binary(16) NOT NULL,
              `key` varchar(50) NOT NULL,
              `value` varchar(2048) NOT NULL,
              `expires_at` datetime NOT NULL,
              `key_id` binary(16) DEFAULT NULL,
              PRIMARY KEY (`id`),
              KEY `session_key_id_fk` (`key_id`),
              CONSTRAINT `session_key_id_fk` FOREIGN KEY (`key_id`) REFERENCES `key` (`id`) ON DELETE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `oauth2_client` (
              `id` binary(16) NOT NULL,
              `name` varchar(150) NOT NULL,
              `secret` varchar(150) NOT NULL,
              `grant_types` varchar(150) DEFAULT NULL,
              `response_types` varchar(150) NOT NULL,
              `redirect_uris` varchar(500) NOT NULL,
              `scope` varchar(150) DEFAULT NULL,
              PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `oauth2_authorization_code` (
              `id` binary(16) NOT NULL,
              `code` varchar(250) NOT NULL,
              `oauth2_client_id` binary(16) NOT NULL,
              `key_id` binary(16) NOT NULL,
              `redirect_uri` varchar(150) NOT NULL,
              `response_type` varchar(150) NOT NULL,
              `scope` varchar(150) DEFAULT NULL,
              `auth_time` datetime NOT NULL,
              `expires_in` bigint DEFAULT NULL,
              `nonce` varchar(150) DEFAULT NULL,
              `code_challenge` varchar(150) NOT NULL,
              `code_challenge_method` varchar(10) DEFAULT NULL,
              PRIMARY KEY (`id`),
              KEY `oauth2_authorization_code_oauth2_client_id_fk` (`oauth2_client_id`),
              KEY `oauth2_authorization_code_key_id_fk` (`key_id`),
              CONSTRAINT `oauth2_authorization_code_key_id_fk` FOREIGN KEY (`key_id`) REFERENCES `key` (`id`) ON DELETE CASCADE,
              CONSTRAINT `oauth2_authorization_code_oauth2_client_id_fk` FOREIGN KEY (`oauth2_client_id`) REFERENCES `oauth2_client` (`id`) ON DELETE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `oauth2_token` (
              `id` binary(16) NOT NULL,
              `access_token` varchar(250) NOT NULL,
              `refresh_token` varchar(250) DEFAULT NULL,
              `scope` varchar(250) DEFAULT NULL,
              `issued_at` DATETIME NOT NULL,
              `access_token_expires_in` INT NOT NULL,
              `refresh_token_expires_in` INT NOT NULL,
              `oauth2_client_id` binary(16) NOT NULL,
              `key_id` binary(16) DEFAULT NULL,
              `token_type` varchar(50) NOT NULL,
              `revoked` enum('Y','N') NOT NULL DEFAULT 'N',
              PRIMARY KEY (`id`),
              KEY `oauth2_token_oauth2_client_id_fk` (`oauth2_client_id`),
              KEY `oauth2_token_key_id_fk` (`key_id`),
              CONSTRAINT `oauth2_token_key_id_fk` FOREIGN KEY (`key_id`) REFERENCES `key` (`id`) ON DELETE CASCADE,
              CONSTRAINT `oauth2_token_oauth2_client_id_fk` FOREIGN KEY (`oauth2_client_id`) REFERENCES `oauth2_client` (`id`) ON DELETE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `acl_role` (
              `id` binary(16) NOT NULL,
              `name` varchar(250) NOT NULL,
              PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `acl_permission` (
              `id` binary(16) NOT NULL,
              `name` varchar(250) NOT NULL,
              PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `acl_role_permission_link` (
              `id` binary(16) NOT NULL,
              `acl_role_id` binary(16) NOT NULL,
              `acl_permission_id` binary(16) NOT NULL,
              KEY `acl_role_permission_link_acl_permission_id_fk` (`acl_permission_id`),
              KEY `acl_role_permission_link_acl_role_id_fk` (`acl_role_id`),
              CONSTRAINT `acl_role_permission_link_acl_permission_id_fk` FOREIGN KEY (`acl_permission_id`) REFERENCES `acl_permission` (`id`) ON DELETE CASCADE,
              CONSTRAINT `acl_role_permission_link_acl_role_id_fk` FOREIGN KEY (`acl_role_id`) REFERENCES `acl_role` (`id`) ON DELETE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
            
            CREATE TABLE `acl_role_key_link` (
              `id` binary(16) NOT NULL,
              `acl_role_id` binary(16) NOT NULL,
              `key_id` binary(16) NOT NULL,
              PRIMARY KEY (`id`),
              KEY `acl_role_key_link_acl_role_id_fk` (`acl_role_id`),
              KEY `acl_role_key_link_key_id_fk` (`key_id`),
              CONSTRAINT `acl_role_key_link_acl_role_id_fk` FOREIGN KEY (`acl_role_id`) REFERENCES `acl_role` (`id`) ON DELETE CASCADE,
              CONSTRAINT `acl_role_key_link_key_id_fk` FOREIGN KEY (`key_id`) REFERENCES `key` (`id`) ON DELETE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;
    
        """
        )

        # inserts the main "Red Warden" key, password "RED_WARDEN"
        model = import_string("red_warden.datazones.RWManager.RWKeyModel")
        key = await model.before_create(
            db,
            {"login": "Red Warden", "enabled": "Y", "password": "RED_WARDEN"},
            skip_validation=True,
        )
        await db.execute(model.insert().values(key))

        # inserts the main OAuth2 client
        model = import_string("red_warden.datazones.RWManager.RWOAuth2ClientModel")
        oauth2_client = await model.before_create(
            db,
            {
                "name": "red_warden_oauth2_client",
                "secret": "RED_WARDEN",
                "grant_types": "authorization_code,password",
                "response_types": "code,token",
                "redirect_uris": "",
                "scope": "manager",
            },
            skip_validation=True,
        )
        await db.execute(model.insert().values(oauth2_client))

        # inserts the global backend
        model = import_string("red_warden.datazones.RWManager.RWBackendModel")
        backend = await model.before_create(
            db,
            {
                "name": "rw_global",
                "destination": RWBackendDestinationEnum.GLOBAL,
                "type": "mysql",
                "params": json.dumps(db.params),
                "endpoints": RWConfig["RW_BACKEND_ENDPOINTS"],
                "enabled": "Y",
            },
            skip_validation=True,
        )
        await db.execute(model.insert().values(backend))


# CREATE TABLE ___RED_WARDEN_MANAGER_DB___.`history` (
#   `id` binary(16) NOT NULL,
#   `record_id` varchar(250) NOT NULL,
#   `who` varchar(250) NOT NULL,
#   `why` enum('INSERT','UPDATE','DELETE') NOT NULL,
#   `what` json NOT NULL,
#   `when` datetime NOT NULL,
#   PRIMARY KEY (`id`)
# ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci;



import json
import base64
import graphene
import httpx
import sqlalchemy
import inspect
import functools
from graphene.utils.module_loading import import_string

_nodes = {}


class RWGraphQLClient:
    _endpoint = None

    def __init__(self, endpoint, **kwargs):
        self._endpoint = endpoint
        super().__init__(**kwargs)

    async def execute(self, payload, **kwargs):
        max_retries = kwargs.pop("max_retries", 5)

        retries = 0
        last_error = None
        while retries < max_retries:
            async with httpx.AsyncClient() as client:
                try:
                    r = await client.post(
                        self._endpoint,
                        json=payload,
                        **kwargs,
                    )
                    r.raise_for_status()
                    return r.json()

                except (httpx.ReadTimeout, httpx.ReadTimeout) as error:
                    last_error = error
                    retries += 1

        raise Exception("Cannot execute query: %s" % last_error)


class RWGraphQLYesNoField(graphene.Enum):
    Yes = "Y"
    No = "N"


class RWGraphQLConnection(graphene.Connection):
    filters = None
    count = graphene.Int()
    total_count = graphene.Int()

    @classmethod
    def __init_subclass_with_meta__(cls, node=None, name=None, **options):
        if isinstance(node, str):
            node = import_string(node)
        setattr(cls, "filters", options.pop("filters", None))
        super().__init_subclass_with_meta__(node=node, name=name, options=options)

    class Meta:
        abstract = True

    @staticmethod
    def resolve_count(root, info, **kwargs):
        return len(root.edges)

    @staticmethod
    def create_field_from_import(class_and_method, filters=None, **kwargs):
        class_path, method_name = class_and_method.rsplit(".", 1)
        connection_class = lambda: import_string(class_path)
        method_name = "resolve_" + method_name
        return RWGraphQLMysqlConnectionField(
            connection_type=connection_class,
            resolver=method_name,
            filters=filters,
            **kwargs,
        )

    @classmethod
    def create_field(cls, method_name, filters=None, **kwargs):
        if not filters and cls.filters:
            filters = cls.filters.get(method_name, None)

        method_name = "resolve_" + method_name
        method = getattr(cls, method_name)
        if not method:
            raise Exception("Unknown method for connection field: %s" % method_name)

        if filters:
            method_filters_class = cls.__name__ + "_" + method_name + "_filters"
            filters_class = type(
                method_filters_class, (graphene.InputObjectType,), filters
            )
            filters = filters_class()

        return RWGraphQLMysqlConnectionField(
            cls, resolver=method, filters=filters, **kwargs
        )


class RWGraphQLNumberFilter(graphene.InputObjectType):
    eq = graphene.Float()
    ne = graphene.Float()
    gt = graphene.Float()
    gte = graphene.Float()
    lt = graphene.Float()
    lte = graphene.Float()
    one_of = graphene.List(graphene.Float)
    not_one_of = graphene.List(graphene.Float)


class RWGraphQLStringFilter(graphene.InputObjectType):
    eq = graphene.String()
    ne = graphene.String()
    like = graphene.String()
    ilike = graphene.String()
    not_like = graphene.String()
    not_ilike = graphene.String()
    one_of = graphene.List(graphene.String)
    not_one_of = graphene.List(graphene.String)


class RWGraphQLMysqlConnectionField(graphene.ConnectionField):
    def __init__(self, connection_type, *args, **kwargs):
        if "args" not in kwargs:
            kwargs["args"] = {}

        resolver = kwargs["resolver"]
        resolver_name = resolver if isinstance(resolver, str) else resolver.__name__
        filters = kwargs.pop("filters", None)
        if not filters and not (
            inspect.isfunction(connection_type)
            or isinstance(connection_type, functools.partial)
        ):
            filters = connection_type.filters.get(resolver_name, None)
            if not filters:
                filters = connection_type.filters.get("default", None)

        if filters:
            method_filters_class = (
                connection_type.__name__ + "_" + resolver_name + "_filters"
            )
            filters_class = type(
                method_filters_class, (graphene.InputObjectType,), filters
            )
            filters = filters_class()
            kwargs["args"]["filters"] = filters

        kwargs["args"]["sort"] = graphene.List(graphene.String)
        super().__init__(connection_type, *args, **kwargs)

    @classmethod
    async def connection_resolver(cls, resolver, connection_type, root, info, **args):
        if isinstance(resolver, str):
            resolver = getattr(connection_type, resolver)

        # the db connection used depends on the datazone of the node's model
        new_args = {
            "db": connection_type._meta.node.model.get_datazone_db(info),
            "filters": args.pop("filters", {}),
            "sort": args.pop("sort", []),
            "pagination": {
                "after": args.pop("after", None),
                "before": args.pop("before", None),
                "first": args.pop("first", 0),
                "last": args.pop("last", 0),
            },
        }

        return super().connection_resolver(
            resolver, connection_type, root, info, **new_args
        )

    @staticmethod
    def get_sort_columns(table, sort_args):
        sort = []
        for s in sort_args:
            if s[0] == "-":
                s = s[1:]
                desc = True
            else:
                desc = False

            if not hasattr(table.c, s):
                raise Exception("Wrong sort field: %s" % s)
            if desc:
                sort.append(table.c[s].desc())
            else:
                sort.append(table.c[s])
        if not sort:
            sort = [table.c.id]
        return sort

    @staticmethod
    def get_filters(table, filters_args):
        filters = []
        for k, v in filters_args.items():
            if not hasattr(table.c, k):
                raise Exception("Wrong filter field: %s" % k)

            if isinstance(v, dict):
                eq = v.get("eq", None)
                ne = v.get("ne", None)
                gt = v.get("gt", None)
                gte = v.get("gte", None)
                lt = v.get("lt", None)
                lte = v.get("lte", None)
                like = v.get("like", None)
                ilike = v.get("ilike", None)
                not_like = v.get("not_like", None)
                not_ilike = v.get("not_ilike", None)
                one_of = v.get("one_of", None)
                not_one_of = v.get("not_one_of", None)

                if eq:
                    filters.append(table.c[k] == eq)
                if ne:
                    filters.append(table.c[k] != ne)
                if lt:
                    filters.append(table.c[k] < lt)
                if lte:
                    filters.append(table.c[k] <= lte)
                if gt:
                    filters.append(table.c[k] > gt)
                if gte:
                    filters.append(table.c[k] >= gte)
                if like:
                    filters.append(table.c[k].like("%" + like + "%"))
                if ilike:
                    filters.append(table.c[k].ilike("%" + ilike + "%"))
                if not_like:
                    filters.append(~table.c[k].like("%" + not_like + "%"))
                if not_ilike:
                    filters.append(~table.c[k].ilike("%" + not_ilike + "%"))
                if one_of:
                    filters.append(table.c[k].in_(one_of))
                if not_one_of:
                    filters.append(~table.c[k].in_(not_one_of))
            else:
                filters.append(table.c[k] == v)

        return filters

    @classmethod
    async def resolve_connection(cls, connection_type, args, resolved):
        pagination = args["pagination"]

        cte_resolved = resolved.cte("resolved")

        limit = int(pagination["first"]) or int(pagination["last"]) or 25
        cursor_hint = pagination["after"] or pagination["before"]
        if cursor_hint:
            try:
                cursor = json.loads(base64.b64decode(cursor_hint.encode()).decode())

            except Exception:
                raise Exception("Wrong cursor format")
            sort = cls.get_sort_columns(cte_resolved, cursor["sort"])
            filters = cls.get_filters(cte_resolved, cursor["filters"])
        else:
            cursor = None
            sort = cls.get_sort_columns(cte_resolved, args["sort"])
            filters = cls.get_filters(cte_resolved, args["filters"])

        cte_ordered = (
            sqlalchemy.select(
                [
                    sqlalchemy.func.row_number().over(order_by=sort).label("n"),
                    *cte_resolved.c,
                ]
            )
            .where(sqlalchemy.and_(*filters))
            .cte("ordered")
        )

        q = cte_ordered.select()
        total_count = await args["db"].fetch_one(
            sqlalchemy.select([sqlalchemy.func.count().label("total")]).select_from(
                q.alias("ordered")
            )
        )

        if cursor:
            cte_id_check = (
                sqlalchemy.select([cte_ordered.c.n]).where(
                    cte_ordered.c.id == bytes.fromhex(cursor["id"])
                )
            ).cte("id_check")

            if pagination["after"]:
                q = q.where(
                    cte_ordered.c.n > sqlalchemy.sql.func.ifnull(cte_id_check.c.n, 0)
                )
            elif pagination["before"]:
                q = q.where(
                    cte_ordered.c.n
                    > sqlalchemy.sql.func.ifnull(
                        sqlalchemy.cast(cte_id_check.c.n, sqlalchemy.BIGINT)
                        - limit
                        - 1,
                        0,
                    )
                )

        if limit:
            q = q.limit(limit + 1)

        def encode_cursor(_node):
            _cursor = {
                "id": "".join("{:02x}".format(b) for b in _node.id),
            }
            if cursor:
                _cursor["filters"] = cursor["filters"]
                _cursor["sort"] = cursor["sort"]
            else:
                _cursor["filters"] = args["filters"]
                _cursor["sort"] = args["sort"]
            return base64.b64encode(json.dumps(_cursor).encode()).decode()

        node_not_included = None
        edges = []
        for index, node in enumerate(await args["db"].fetch_all(q)):
            if index < limit:
                edges.append(
                    connection_type.Edge(node=node, cursor=encode_cursor(node))
                )
            else:
                node_not_included = node

        if edges:
            start_cursor = edges[0].cursor
            end_cursor = edges[-1].cursor
            has_previous_page = edges[0].node.n > 1
            has_next_page = node_not_included is not None
        else:
            start_cursor = None
            end_cursor = None
            has_previous_page = False
            has_next_page = False

        return connection_type(
            total_count=total_count["total"],
            edges=edges,
            page_info=graphene.PageInfo(
                start_cursor=start_cursor,
                end_cursor=end_cursor,
                has_previous_page=has_previous_page,
                has_next_page=has_next_page,
            ),
        )


class RWGraphQLNode(graphene.Node):
    class Meta:
        name = "Node"

    @classmethod
    def to_global_id(cls, _type, _id):
        str_id = "".join("{:02x}".format(b) for b in _id)
        return base64.b64encode(f"{_type + ':' + str_id}".encode()).decode()

    @classmethod
    async def get_node_from_global_id(
        cls, info, global_id, only_type=None, ignore_permissions=False
    ):
        _type, _id = base64.b64decode(global_id.encode()).decode().split(":")
        if only_type:
            # We assure that the node type that we want to retrieve
            # is the same that was indicated in the field type
            if not isinstance(only_type, str):
                only_type = only_type._meta.name

            if _type != only_type:
                raise Exception("Received not compatible node.")

        node_class = _nodes.get(_type, None)
        if not node_class:
            raise Exception("Node type %s does not have an associated class" % _type)
        return await node_class.get_node(info, bytes.fromhex(_id), ignore_permissions)


class RWGraphQLObjectType(graphene.ObjectType):
    model = None

    def __init_subclass__(cls, **kwargs):
        _nodes[cls.__name__] = cls
        super().__init_subclass__(**kwargs)

    @classmethod
    async def check_permissions(cls, info, node):
        raise NotImplementedError("ERROR_PERMISSIONS_NOT_SET_FOR_OBJECT")

    @classmethod
    async def get_node(cls, info, _id, ignore_permissions=False):
        if not cls.model:
            raise Exception("Missing model information for %s" % cls.__name__)

        db = cls.model.get_datazone_db(info)
        data = await db.fetch_one(cls.model.select().where(cls.model.c.id == _id))
        if data:
            node = cls()
            for k, v in data.items():
                if k in node._meta.fields.keys():
                    setattr(node, k, v)

            if ignore_permissions:
                return node
            else:
                return await cls.check_permissions(info, node)


def create_query_container(query_containers):
    if not isinstance(query_containers, list):
        query_containers = [query_containers]

    queries = {}
    for qc in query_containers:
        queries.update(qc)
    queries["node"] = RWGraphQLNode.Field()

    class_name = "graphql_query"
    the_class = type(class_name, (graphene.ObjectType,), queries)
    return the_class


def create_mutation_container(mutation_containers):
    if not isinstance(mutation_containers, list):
        mutation_containers = [mutation_containers]

    mutations = {}
    for mc in mutation_containers:
        mutations.update(mc)
    class_name = "graphql_mutation"
    the_class = type(class_name, (graphene.ObjectType,), mutations)
    return the_class



from red_warden.mvc import RWValidationModelBase
from red_warden.enums import RWYesNoEnum
from red_warden.datalayer import RWDataLayer
from red_warden.datazones.RWManager.models import RWKeyModel


class RWSessionLogic:
    @classmethod
    async def authenticate(cls, **kwargs):
        class ValidationModel(RWValidationModelBase):
            username: str
            password: str

        data = ValidationModel(**kwargs).dict()

        row = await RWDataLayer.load_one(
            RWKeyModel, filters=["login:eq:%s" % data["username"]]
        )
        if not row or not RWKeyModel.check_password(
            row["password_hash"], data["password"]
        ):
            return False

        return row

    @classmethod
    async def toggle_mfa(cls, **kwargs):
        class ValidationModel(RWValidationModelBase):
            key_id: str
            status: RWYesNoEnum

        data = ValidationModel(**kwargs).dict()

        row = await RWDataLayer.load_by_id(RWKeyModel, data["key_id"], None)
        if not row:
            return None

        return await RWDataLayer.save(
            RWKeyModel,
            data["key_id"],
            RWKeyModel.toggle_mfa(row["mfa_enabled"], data["status"]),
            None,
            True,
        )

    @classmethod
    async def check_otp(cls, **kwargs):
        class ValidationModel(RWValidationModelBase):
            key_id: str
            otp_code: str

        data = ValidationModel(**kwargs).dict()

        row = await RWDataLayer.load_by_id(RWKeyModel, data["key_id"], None)
        if not row:
            return None

        if not RWKeyModel.check_otp(row["mfa_secret"], data["otp_code"]):
            return False

        return row

    @classmethod
    async def login(cls, logged_in_key):
        # generic scope used when a function needs a user but not a specific scope
        scope = ["logged_in"]

        return {"key_id": logged_in_key["id"], "scope": scope}



import graphene
from red_warden.graphql import (
    RWGraphQLConnection,
    RWGraphQLStringFilter,
    RWGraphQLYesNoField,
)
from red_warden.datazones.RWManager import KeyModel
from red_warden.datazones.graphql.key.types import Key


class KeyConnection(RWGraphQLConnection):
    class Meta:
        node = Key
        filters = {
            "default": {
                "login": RWGraphQLStringFilter(),
                "enabled": RWGraphQLYesNoField(),
            }
        }

    @staticmethod
    def resolve_all(_, info, **args):
        return KeyModel.select()


class TenantKeyByTenantConnectionFilters(graphene.InputObjectType):
    tenant_id = RWGraphQLStringFilter(required=True)



import json
import typing
from starlette.responses import Response
from red_warden.helpers import RWCustomJsonEncoder


class RWJSONResponse(Response):
    media_type = "application/json"

    def render(self, content: typing.Any) -> bytes:
        return json.dumps(
            content,
            cls=RWCustomJsonEncoder,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
        ).encode("utf-8")



import sqlalchemy
from red_warden.graphql import (
    RWGraphQLConnection,
    RWGraphQLStringFilter,
    RWGraphQLYesNoField,
)
from red_warden.datazones.RWManager import TenantModel
from red_warden.datazones.graphql.tenant.types import Tenant


class TenantConnection(RWGraphQLConnection):
    class Meta:
        node = Tenant
        filters = {
            "default": {
                "name": RWGraphQLStringFilter(),
                "enabled": RWGraphQLYesNoField(),
            }
        }

    @staticmethod
    def resolve_all(_, info, **args):
        cte_non_recursive = (
            sqlalchemy.select(
                [
                    TenantModel.c.id,
                    TenantModel.c.name,
                    TenantModel.c.params,
                    TenantModel.c.enabled,
                    TenantModel.c.parent_id,
                    sqlalchemy.literal(1).label("level"),
                    TenantModel.c.name.label("path"),
                ]
            )
            .where(TenantModel.c.parent_id == None)
            .cte("trees", recursive=True)
        )

        cte_recursive = sqlalchemy.select(
            [
                TenantModel.c.id,
                TenantModel.c.name,
                TenantModel.c.params,
                TenantModel.c.enabled,
                TenantModel.c.parent_id,
                cte_non_recursive.c.level + 1,
                cte_non_recursive.c.path
                + sqlalchemy.literal(" > ")
                + TenantModel.c.name,
            ]
        ).where(TenantModel.c.parent_id == cte_non_recursive.c.id)

        return cte_non_recursive.union_all(cte_recursive).select()



from red_warden.datalayer import RWDatazone
from .models import (
    RWBackendModel,
    RWTenantModel,
    RWDatapathModel,
    RWKeyModel,
    RWSessionModel,
    RWOAuth2ClientModel,
    RWOAuth2AuthorizationCodeModel,
    RWOAuth2TokenModel,
    RWAclRoleModel,
    RWAclPermissionModel,
    RWAclRolePermissionLinkModel,
)
from .controllers import (
    RWBackendController,
    RWTenantController,
    RWKeyController,
    RWSessionController,
    RWOAuth2ClientController,
    RWOAuth2AuthorizationCodeController,
    RWOAuth2TokenController,
    RWOAuth2FlowController,
    RWAclRoleController,
    RWAclPermissionController,
)


class RWManager(RWDatazone):
    models = [
        RWBackendModel,
        RWTenantModel,
        RWDatapathModel,
        RWKeyModel,
        RWSessionModel,
        RWOAuth2ClientModel,
        RWOAuth2AuthorizationCodeModel,
        RWOAuth2TokenModel,
        RWAclRoleModel,
        RWAclPermissionModel,
        RWAclRolePermissionLinkModel,
    ]
    controllers = [
        RWBackendController,
        RWTenantController,
        # RWDatapathController,
        RWKeyController,
        RWSessionController,
        RWOAuth2ClientController,
        RWOAuth2AuthorizationCodeController,
        RWOAuth2TokenController,
        RWOAuth2FlowController,
        RWAclRoleController,
        RWAclPermissionController,
    ]


# class RWMysqlHistoryModel(RWMysqlModel):
#     datazone = "history"
#
#     @staticmethod
#     def get_db_columns():
#         return [
#             sqlalchemy.Column(
#                 "table_record_id", sqlalchemy.String(length=250), nullable=False
#             ),
#             sqlalchemy.Column("who", sqlalchemy.String(length=250), nullable=False),
#             sqlalchemy.Column(
#                 "why", sqlalchemy.Enum("INSERT", "UPDATE", "DELETE"), nullable=False
#             ),
#             sqlalchemy.Column("what", sqlalchemy.JSON(), nullable=False),
#             sqlalchemy.Column("when", sqlalchemy.DateTime(), nullable=False),
#             sqlalchemy.Column("description", sqlalchemy.String(length=500)),
#         ]



import graphene
from red_warden.graphql import (
    RWGraphQLNode,
    RWGraphQLConnection,
    RWGraphQLObjectType,
    RWGraphQLYesNoField,
)
from red_warden.datazones.RWManager import TenantModel


class Tenant(RWGraphQLObjectType):
    """A tenant, used to manage the accessed data"""

    @classmethod
    async def check_permissions(cls, info, node):
        return node

    model = TenantModel

    class Meta:
        interfaces = (RWGraphQLNode,)

    name = graphene.String(description="The tenant name")
    parent_id = graphene.String(description="The tenant parent id")
    password = graphene.String(description="The tenant password, encrypted")
    params = graphene.JSONString(description="The tenant params")
    enabled = RWGraphQLYesNoField(description="Is tenant enabled?")
    level = graphene.Int(description="The tenant nesting level")
    children = graphene.List("schemas.tenant.types.Tenant")
    # datapaths = RWGraphQLConnection.create_field_from_import(
    #     "schemas.link_tenant_datapath.connections.TenantDatapathsConnection.all"
    # )
    # keys = RWGraphQLConnection.create_field_from_import(
    #     "schemas.link_tenant_key.connections.TenantKeysConnection.all"
    # )

    @staticmethod
    async def resolve_children(parent, info, **args):
        db = TenantModel.get_datazone_db(info)
        return await db.fetch_all(
            TenantModel.select().where(TenantModel.c.parent_id == parent.id)
        )



import os
import graphene
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.backends import default_backend
from cryptography.exceptions import InvalidKey
from red_warden.graphql import RWGraphQLYesNoField
from red_warden.datazones.RWManager import KeyModel
from red_warden.datazones.graphql.key.types import Key


class CreateKey(graphene.ClientIDMutation):
    key = graphene.Field(Key, required=True)
    mfa_secret = graphene.String()

    class Input:
        login = graphene.String(required=True)
        password = graphene.String(required=True)
        enabled = RWGraphQLYesNoField(required=True)
        mfa_enabled = RWGraphQLYesNoField()

    @classmethod
    async def mutate_and_get_payload(cls, root, info, **kwargs):
        db = KeyModel.get_datazone_db(info)

        async with db.transaction() as tx:
            kwargs["id"] = await db.get_uuid()

            password = kwargs.pop("password", None)
            if password:
                salt = os.urandom(32)
                kdf = Scrypt(
                    salt=salt, length=32, n=2 ** 14, r=8, p=1, backend=default_backend()
                )
                b = kdf.derive(password.encode())
                kwargs["password_hash"] = "%s.%s" % (salt.hex(), b.hex())

            if kwargs["mfa_enabled"] == "Y":
                kwargs["mfa_secret"] = Fernet.generate_key()

            mfa_secret = kwargs.get("mfa_secret", None)

            await db.execute(KeyModel.insert().values(kwargs))

            kwargs.pop("password_hash", None)
            kwargs.pop("mfa_secret", None)
            return cls(key=Key(**kwargs), mfa_secret=mfa_secret.decode())


class CheckPassword(graphene.ClientIDMutation):
    key = graphene.Field(Key, required=True)

    class Input:
        login = graphene.String(required=True)
        password = graphene.String(required=True)

    @classmethod
    async def mutate_and_get_payload(cls, root, info, **kwargs):
        db = KeyModel.get_datazone_db(info)

        async with db.transaction():
            try:
                k = await db.fetch_one(
                    KeyModel.select().where(KeyModel.c.login == kwargs["login"])
                )

                salt, key = k.password_hash.split(".")
                kdf = Scrypt(
                    salt=bytes.fromhex(salt),
                    length=32,
                    n=2 ** 14,
                    r=8,
                    p=1,
                    backend=default_backend(),
                )
                kdf.verify(kwargs["password"].encode(), bytes.fromhex(key))

                key = Key.get_node(info, k.id)

            except InvalidKey:
                raise Exception("Invalid credentials")

        return cls(key=key)



import graphene
from red_warden.graphql import RWGraphQLNode
from models import DatapathModel
from schemas.datapath.types import Datapath


class CreateDatapath(graphene.ClientIDMutation):
    datapath = graphene.Field(Datapath, required=True)

    class Input:
        backend_id = graphene.String()
        name = graphene.String(required=True)
        params = graphene.JSONString()

    @classmethod
    async def mutate_and_get_payload(cls, root, info, **kwargs):
        db = DatapathModel.get_datazone_db(info)

        async with db.transaction():
            kwargs["id"] = await db.get_uuid()

            backend_id = kwargs.pop("backend_id", None)
            if backend_id:
                node = await RWGraphQLNode.get_node_from_global_id(
                    info, backend_id, "Backend", ignore_permissions=True
                )
                kwargs["backend_id"] = node.id

            params = kwargs.get("params", None)
            if not params:
                kwargs["params"] = {}

            await db.execute(DatapathModel.insert().values(kwargs))

            kwargs.pop("backend_id", None)
            return cls(datapath=Datapath(**kwargs))



from abc import ABC
import sqlalchemy
import json
import pydantic
from typing import Optional
from cryptography.fernet import Fernet
from mako.lookup import TemplateLookup
from mako.exceptions import TemplateLookupException
from starlette.exceptions import HTTPException
from starlette.responses import HTMLResponse
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from red_warden.config import RWConfig
from red_warden.datalayer import RWDataLayer


# region --- Models -------------------------------------------------------------------------------------------------
# Database table definitions.
metadata = sqlalchemy.MetaData()


class RWMysqlModelMeta(type):
    table = None

    def __getattr__(self, item):
        return getattr(self.table, item)


class RWValidationModelBase(pydantic.BaseModel):
    class Config:
        extra = "forbid"


class RWOptionalValidationModelMetaclass(pydantic.main.ModelMetaclass):
    # https://stackoverflow.com/questions/67699451/make-every-fields-as-optional-with-pydantic
    def __new__(cls, name, bases, namespaces, **kwargs):
        annotations = namespaces.get("__annotations__", {})
        for base in bases:
            annotations = {**annotations, **base.__annotations__}
        for field in annotations:
            if not field.startswith("__"):
                annotations[field] = Optional[annotations[field]]
        namespaces["__annotations__"] = annotations
        return super().__new__(cls, name, bases, namespaces, **kwargs)


class RWMysqlModel(metaclass=RWMysqlModelMeta):
    table_name = None

    class ValidationModel(RWValidationModelBase):
        pass

    def __init_subclass__(cls):
        columns = [
            sqlalchemy.Column("id", sqlalchemy.Binary(16), primary_key=True),
        ] + cls.get_db_columns()
        cls.table = sqlalchemy.Table(cls.table_name, metadata, *columns)

    @staticmethod
    def get_db_columns():
        raise NotImplementedError

    @staticmethod
    def get_secret():
        return str(RWConfig["RW_DATABASE_SECRET"])

    @classmethod
    def get_special_columns(cls):
        return {"id": ["id"], "encrypted": [], "json": []}

    @classmethod
    def get_select_query(cls):
        return cls.select()

    @classmethod
    def _encrypt_columns(cls, row):
        sc = cls.get_special_columns()

        for ec in sc["encrypted"]:
            if ec in row.keys() and row[ec]:
                fernet = Fernet(cls.get_secret())
                row[ec] = fernet.encrypt(row[ec].encode()).decode("utf-8")

    @classmethod
    def _decrypt_columns(cls, row):
        sc = cls.get_special_columns()

        for ec in sc["encrypted"]:
            if ec in row.keys() and row[ec]:
                fernet = Fernet(cls.get_secret())
                row[ec] = fernet.decrypt(row[ec].encode("utf-8")).decode("utf-8")

    @classmethod
    def _convert_id_columns_to_string(cls, row):
        sc = cls.get_special_columns()

        for ic in sc["id"]:
            if ic in row.keys() and row[ic]:
                row[ic] = bytes.hex(row[ic]).upper()

    @classmethod
    def _convert_id_columns_to_byte_array(cls, row):
        sc = cls.get_special_columns()

        for ic in sc["id"]:
            if ic in row.keys() and row[ic]:
                row[ic] = bytes.fromhex(row[ic])

    @classmethod
    def _convert_json_columns_to_object(cls, row):
        sc = cls.get_special_columns()
        for jc in sc["json"]:
            if jc in row.keys() and row[jc]:
                row[jc] = json.loads(row[jc])

    @classmethod
    def _convert_json_columns_to_string(cls, row):
        sc = cls.get_special_columns()
        for jc in sc["json"]:
            if jc in row.keys() and row[jc] and not isinstance(row[jc], str):
                row[jc] = json.dumps(row[jc])

    @classmethod
    async def after_load(cls, db, row):
        row = dict(row)
        cls._decrypt_columns(row)
        cls._convert_id_columns_to_string(row)
        cls._convert_json_columns_to_object(row)
        return row

    @classmethod
    async def before_create(cls, db, row, skip_validation=False):
        row = dict(row)
        cls._convert_json_columns_to_string(row)

        if not skip_validation:
            row = cls.ValidationModel(**dict(row)).dict()

        row["id"] = await db.get_uuid()
        cls._convert_id_columns_to_byte_array(row)
        cls._encrypt_columns(row)
        return row

    @classmethod
    async def before_update(cls, db, row, skip_validation=False):
        row = dict(row)
        cls._convert_json_columns_to_string(row)

        # validates only the fields passed to the update method
        if not skip_validation:

            class ValidationModelUpdate(
                cls.ValidationModel, metaclass=RWOptionalValidationModelMetaclass
            ):
                pass

            validated_row = ValidationModelUpdate(**row).dict()
            for k in row.keys():
                row[k] = validated_row[k]

        cls._convert_id_columns_to_byte_array(row)
        cls._encrypt_columns(row)
        return row

    @classmethod
    async def after_create(cls, db, row):
        return row

    @classmethod
    async def after_update(cls, db, row):
        return row


# endregion ---------------------------------------------------------------------------------------------------------


# region --- Controllers --------------------------------------------------------------------------------------------
class RWController(ABC):
    custom_view_data = {}

    @classmethod
    def get_routes(cls):
        raise NotImplementedError

    @classmethod
    def add_custom_view_data(cls, custom_view_data):
        cls.custom_view_data.update(custom_view_data)

    @classmethod
    def _prepare_view_data(cls, request, view_data=None):
        if not view_data:
            view_data = {}

        def url_for(route_name, locale=None, **path_params):
            if hasattr(request.state, "i18n") and not locale:
                locale = request.state.i18n.locale

            url = None
            if locale:
                try:
                    url = request.url_for("%s_%s" % (route_name, locale), **path_params)
                except:
                    pass

            if not url:
                url = request.url_for(route_name, **path_params)

            return url

        view_data.update(
            {
                "config": RWConfig,
                "url_for": url_for,
            }
        )

        if hasattr(request.state, "i18n"):
            view_data.update({"i18n": request.state.i18n})

        if hasattr(request.state, "session"):
            view_data.update({"session": request.state.session})

        if hasattr(request.state, "identity"):
            view_data.update({"identity": request.state.identity})

        view_data.update(cls.custom_view_data)

        return view_data

    @classmethod
    def render_view(cls, name, request, view_data=None):
        return HTMLResponse(
            RWView.render(name, cls._prepare_view_data(request, view_data))
        )


class RWRestController(RWController, ABC):
    model = None

    def __init_subclass__(cls, **kwargs):
        if not cls.model:
            raise Exception("[%s] Model cannot be empty" % cls.__name__)

    @classmethod
    async def load_by_id(cls, request):
        model_id = request.path_params.get("id", None)

        pagination = None
        data = await RWDataLayer.load_by_id(cls.model, model_id, request.state.identity)
        if not data:
            raise HTTPException(404)

        results = {"data": data}
        if pagination:
            results["pagination"] = pagination

        return results

    @classmethod
    async def load_all(cls, request):
        pagination = {
            "after": request.query_params.get("after", None),
            "before": request.query_params.get("before", None),
            "limit": request.query_params.get("limit", 0),
        }

        sort = request.query_params.get("sort", None)
        if sort:
            sort = sort.split(",")

        filters = request.query_params.get("filters", None)
        if filters:
            filters = filters.split(",")
        else:
            filters = []

        data, pagination = await RWDataLayer.load_all(
            cls.model, request.state.identity, pagination, sort, filters
        )

        results = {"data": data}
        if pagination:
            results["pagination"] = pagination

        return results

    @classmethod
    async def create(cls, request):
        return {
            "data": await RWDataLayer.save(
                cls.model, None, await request.form(), request.state.identity
            )
        }

    @classmethod
    async def update(cls, request):
        data = await RWDataLayer.save(
            cls.model,
            request.path_params.get("id", None),
            await request.form(),
            request.state.identity,
        )
        if data:
            return {"data": data}
        else:
            raise HTTPException(404)

    @classmethod
    async def delete(cls, request):
        model_id = request.path_params.get("id", None)
        data = await RWDataLayer.remove(cls.model, model_id, request.state.identity)
        if data:
            return {"data": data}
        else:
            raise HTTPException(404)


class RWStaticFilesController(RWController):
    project_dir = None
    output_dir = None

    def __init__(self, project_dir, output_dir):
        self.project_dir = project_dir
        self.output_dir = output_dir

    def get_routes(self):
        return [
            Mount(
                self.output_dir,
                app=StaticFiles(directory=self.project_dir, html=True),
                name="static",
            )
        ]


# class RWGraphQLController(RWController):
#     _prefix = None
#     _schema = None
#
#     # TODO setup fields
#     # def __init__(self, query, mutation, prefix="/api"):
#     #     self._prefix = prefix
#     #     self._schema = graphene.Schema(query, mutation, auto_camelcase=False)
#     #     super().__init__()
#
#     @classmethod
#     def get_routes(cls):
#         routes = [
#             Route(
#                 cls._prefix,
#                 GraphQLApp(schema=cls._schema, executor_class=AsyncioExecutor),
#                 methods=["POST"],
#                 name="graphql",
#             ),
#         ]
#
#         return routes


# endregion ------------------------------------------------------------------------------------------------------------


# region --- Views --------------------------------------------------------------------------------------------------
class RWView:
    template_lookup = None

    @classmethod
    def render(cls, name, data=None):
        if not cls.template_lookup:
            dirs = RWConfig["RW_VIEWS_TEMPLATE_DIR"]
            if isinstance(dirs, str):
                dirs = dirs.split(",")

            cls.template_lookup = TemplateLookup(directories=[*dirs])

        try:
            mytemplate = cls.template_lookup.get_template("%s.mako" % name)
        except:
            if "i18n" in data.keys():
                mytemplate = cls.template_lookup.get_template(
                    "%s/%s.mako" % (data["i18n"].locale, name)
                )
            else:
                raise TemplateLookupException("View %s not found" % name)

        return mytemplate.render(**(data or {}))


# endregion ------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['red_warden',
 'red_warden.__engines',
 'red_warden.__graphql',
 'red_warden.__graphql.__datazones.graphql.datapath',
 'red_warden.__graphql.__datazones.graphql.key',
 'red_warden.__graphql.__datazones.graphql.tenant',
 'red_warden.__graphql.__datazones.rw_manager',
 'red_warden.datazones.RWManager',
 'red_warden.datazones.RWManager.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Mako>=1.1.4,<2.0.0',
 'SQLAlchemy<1.4',
 'aioauth>=1.0.3,<2.0.0',
 'aioredis>=1.3.1,<2.0.0',
 'cryptography>=3.4.7,<4.0.0',
 'databases[mysql]>=0.4.3,<0.5.0',
 'gunicorn>=20.1.0,<21.0.0',
 'httpx>=0.18.1,<0.19.0',
 'motor>=2.4.0,<3.0.0',
 'pydantic[email]>=1.8.2,<2.0.0',
 'pyotp>=2.6.0,<3.0.0',
 'starlette[full]>=0.14.2,<0.15.0',
 'uvicorn[standard]>=0.13.4,<0.14.0']

setup_kwargs = {
    'name': 'red-warden',
    'version': '0.4.3',
    'description': 'Resilient, Extensible and Distributed Web Application Rapid Development ENgine',
    'long_description': None,
    'author': "Daniele 'px' Picca (Sevenia)",
    'author_email': 'd.picca@sevenia.tech',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
